"""Configuration variables

Author:
    (C) Marvin Mager - @mvnmgrx - 2022

License identifier:
    GPL-3.0

Major changes:
    19.09.2022 - created
"""

KIUTILS_CREATE_NEW_VERSION_STR = '20211014'
"""Version string used in ``create_new()`` class functions"""

KIUTILS_CREATE_NEW_GENERATOR_STR = 'kiutils'
"""Generator string used in ``create_new()`` class functions"""